local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer
local lms = love.mouse

local winObj = {}
function winObj:create(v)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(v)
	return _i
end
function winObj:setup(v)
	v = v or {}
	self.x = v.x or 0
	self.y = v.y or 0
	self.w = v.w or 150
	self.h = v.h or 100

	self.conW = self.w - 2
	self.conH = self.h - 12

	self.parent = v.parent or windowingUIState

	self.lastTimeFocused = tmr.getTime()

	self.contextCanvas = lgr.newCanvas(self.conW, self.conH, {dpiscale = 1})

	local c = v.app
	if not apps[c] then c = "sample" end
	self.context = apps[c]:create { w = self.conW, h = self.conH, args = v.args, parent = self }

	self.resizable = not self.context.noResize

	self.id = v.id
	self.nameID = v.nameID

	self.title = self.context.defaultTitle or "Window " .. self.id
	-- print(">", self.title)
	self:genTitleBar()
end

function winObj:update(dt)
	if self.adjust then
		self:resize()
		self.adjust = false
	end

	self.mouseX = math_clamp(stateStack.mouseX - self.x - 1, 0, self.w - 2)
	self.mouseY = math_clamp(stateStack.mouseY - self.y - 11, 0, self.h - 12)
	if self.context.updateInBackground or self.focused then
		self.context:update(dt)
	end

	if self.isResizing then
		self:resize()
	end

	if self.minimized then return false end
	self.contextCanvas:renderTo(function()
		lgr.clear(self.context.bgColor)
		self.context:draw()
	end)
end

function winObj:draw()
	local MYHEIGHT = not self.minimized and self.h or 12
	lgr.setFont(font.dialog)
	lgr.push()
	lgr.translate(self.x, self.y)

	-- Window Shadow
	if self.focused then
		lgr.setColor(0,0,0,0.4)
		lgr.rectangle("fill", 1 * (self.focused and 2 or 1), 1 * (self.focused and 2 or 1), self.w, MYHEIGHT)
	end

	-- Window Border
	lgr.setColor(0,0,0)
	if not self.focused then
		lgr.setColor(0.25,0.25,0.3)
	end
	lgr.rectangle("fill", 0, 0, self.w, MYHEIGHT)

	-- Window Internal
	lgr.setColor(1,1,1)
	lgr.rectangle("fill", 1, 1, self.w - 2, MYHEIGHT - 2)

	-- Window Title
	lgr.setColor(1,1,1)
	lgr.draw(self.titleBar, 1, 1)

	if not self.minimized then
		lgr.translate(1, 11)
		lgr.setColor(1,1,1)
		lgr.draw(self.contextCanvas, 0, 0)

		-- Debug Mouse Coordinates
		if self.focused then
			if kbd.isDown("\\") then
				lgr.setColor(0,0,0)
				lgr.print(self.mouseX .. ", " .. self.mouseY, 2, 2)
			end
		end

		if self.focused and self.resizable and overlap(self.mouseX, self.mouseY, 1, 1, 0, 0, self.w, self.h) then
			lgr.push()
			lgr.translate(self.w - 11, self.h - 11 - 10)
			lgr.setColor(1,1,1,0.5)
			lgr.polygon("fill", 0, 8, 8, 8, 8, 0)
			lgr.setColor(0,0,0,0.5)
			lgr.polygon("line", 0, 8, 8, 8, 8, 0)
			lgr.pop()
		end
	end
	lgr.pop()
end

function winObj:mousepressed(x, y, button)
	local x2 = x - self.x
	local y2 = y - self.y
	if overlap(x2, y2, 1, 1, 0, 0, self.w, 11) then
		for i, b in pairs(self.titleButtons) do
			-- print("Button:", i, b.symbol)
			-- print("Window >>", b.x, b.y, x2, y2)
			if b:mousepressed(x2, y2, button) then return true end
		end

		self.beingDragged = true
		self.mousePressOffsetX = x - self.x
		self.mousePressOffsetY = y - self.y
		-- print(self.mousePressOffsetX, self.mousePressOffsetY)
		return
	elseif overlap(x2, y2, 1, 1, self.w - 10, self.h - 10, self.w, self.h) and self.resizable then
		self.beingResized = true
		self.resizeOX = (x - self.x) - (self.w - 10)
		self.resizeOY = (y - self.y) - (self.h - 10)

		self.oldWidth = self.w
		self.oldHeight = self.h
		-- print(self.resizeOX, self.resizeOY)
		return
	end
	if self.context.mousepressed then
		self.context:mousepressed(self.mouseX, self.mouseY, button)
	end
end

function winObj:mousereleased(x, y, button)
	local x2 = x - self.x
	local y2 = y - self.y
	for i, b in pairs(self.titleButtons) do
		if b:mousereleased(x2, y2, button) then return true end
	end

	self.beingDragged = false
	if self.beingResized and (self.w ~= self.oldWidth) and (self.h ~= self.oldHeight) then
		self.beingResized = false
		self.isResizing = false
		self:resize()
	end

	if self.context.mousereleased then
		self.context:mousereleased(self.mouseX, self.mouseY, button)
	end
end

function winObj:mousemoved(x, y, dx, dy)
	if self.beingDragged then
		self.x = x - self.mousePressOffsetX
		self.y = math_clamp(y - self.mousePressOffsetY, 10, stateStack.drawHeight - 12)
	elseif self.beingResized then
		self.isResizing = true
		self.w = math.max(x - self.x + self.resizeOX, 62)
		self.h = math.max(y - self.y + self.resizeOY, 62)
	end

	if self.context.mousemoved then
		self.context:mousemoved(self.mouseX, self.mouseY, dx, dy)
	end
end

function winObj:doubleclick(x, y, button)
	local x2 = x - self.x
	local y2 = y - self.y
	if overlap(x2, y2, 1, 1, 0, 0, self.w, 11) and not self.context.noMinimize then
		self.parent:maximizeWindow(self.nameID)
		self.beingDragged = false
	end
	-- self.x = 0
	-- self.y = 0
	-- self.w = stateStack.drawWidth
	-- self.h = stateStack.drawHeight
	if self.context.doubleclick then
		self.context:doubleclick(self.mouseX, self.mouseY, button)
	end
end

function winObj:wheelmoved(x, y)
	print(x, y)
end

function winObj:focus()
	self:genTitleBar()
end

function winObj:blur()
	self:genTitleBar()
end

function winObj:resize()
	self.w = math.floor(self.w)
	self.h = math.floor(self.h)

	self.conW = self.w - 2
	self.conH = self.h - 12

	self.contextCanvas = lgr.newCanvas(self.conW, self.conH, {dpiscale = 1})
	self.context:resize(self.conW, self.conH)
	self:genTitleBar()
end

function winObj:genTitleBar()
	self.titleButtons = {}
	self.titleButtons[1] = genButtonObj:create { x = 2, y = 2, symbol = "close", callback = function() self.parent:closeWindow(self.nameID) end }
	if not self.context.noMinimize then
		self.titleButtons[2] = genButtonObj:create { x = self.w - 9, y = 2, symbol = "minimize", callback = function() self.parent:minimizeWindow(self.nameID) end }
	end
	-- self.titleButtons[3] = genButtonObj:create { x = 1, y = 1, symbol = "close" }


	self.titleBar = lgr.newCanvas(self.w - 2, 10)
	self.titleBar:renderTo(function()
		local w = font.dialog:getWidth(self.title)
		local x = math.floor((self.w / 2) - (w / 2))
		lgr.clear(0,0,0,0)

		lgr.setColor(0.4,0.4,0.8)
		if not self.focused then
			lgr.setColor(0.35,0.35,0.4)
		end
		lgr.rectangle("fill", 0, 0, self.w - 2, 10)

		if self.focused then
			lgr.setColor(0,0,0,0.25)
			for y = 1,4 do
				local ny = (y * 2)
				lgr.line(4, ny, self.w - 4, ny)
			end
			lgr.line(-1, 10, self.w, 10)
		end

		lgr.setFont(font.dialog)

		lgr.setColor(0.4,0.4,0.8)
		if not self.focused then
			lgr.setColor(0.35,0.35,0.4)
		end
		lgr.rectangle("fill", x - 2, 0, w + 4, 9)

		lgr.setColor(1,1,1)
		if not self.focused then
			lgr.setColor(0.7,0.7,0.8)
		end
		lgr.print(self.title, x, 1)

		if self.focused then
			for i, b in pairs(self.titleButtons) do
				lgr.setColor(0.4,0.4,0.8)
				lgr.rectangle("fill", b.x - 2, b.y - 1, b.w + 3, b.h + 2)
				lgr.setColor(1,1,1)
				b:draw()
			end
		end
	end)
end

return winObj
