local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer

local arg = ... or {}

local scriptModule = {}

local _modules = lfs.getDirectoryItems("scriptmodules")
for i, f in pairs(_modules) do
	if f:sub(-3) == "lua" then
		local file = string.lower(f:sub(1, -5))
		scriptModule[file] = lfs.load("scriptmodules/"..f)()
	end
end

function adjustIndentation(s)
	local ind = 0
	for i = 1, #s.lines do
		local command = s.lines[i].command
		s.lines[i].indent = ind
		if command == "ENDIF" then
			ind = ind - 1
		elseif command == "CHECKIF" or command == "IFANSWER" or command == "CHECKRANDOM" then
			ind = ind + 1
		end
	end
end

function adjustEndIfs(s)
	for i = 1, #s.lines do
		local sl = s.lines[i]
		if sl.command == "CHECKIF" or sl.command == "IFANSWER" or sl.command == "CHECKRANDOM" then
			local foundOne = false
			for j = i, #s.lines do
				if s.lines[j].command == "ENDIF" then
					foundOne = true
					sl.endIf = j
					break
				end
			end
		end
	end
end

function adjustJumpExits(s)
	for i = 1, #s.lines do
		local sl = s.lines[i]
		if sl.command == "JUMPEXIT" then
			local jump = string.sub(trim(sl.rawLine), 3, -1)
			s.jumps[jump] = i
			sl.jumpName = sl.rawLine
		end
	end
end

function processScriptFile(fileData)
	local s = {}
	s.lines = {}
	s.jumps = {}
	local i = 0
	for z, line in pairs(fileData) do
		i = i + 1
		local _dontWait = false
		if string.find(line, "dontwait") then
			_dontWait = true
			local nl, _ = string.gsub(line, "dontwait", "")
			line = nl
			-- print("Not going to wait", i)
		end
		local _locOfSpace = string.find(line, " ")
		local command = line
		local args = {}
		if _locOfSpace then
			command = string.upper(trim(string.sub(line, 1, _locOfSpace - 1)))
			args = {string.sub(line, _locOfSpace + 1, -1)}
		end
		if string.find(line, ">>") then
			command = "JUMPEXIT"
			args = {}
		end
		s.lines[i] = { command = command, args = args, rawLine = line, indent = 0, completed = 0, dontWait = dontWait }
	end

	adjustJumpExits(s)
	adjustIndentation(s)
	adjustEndIfs(s)

	return s
end

function loadScript(file)
	-- if file == "" then return false end
	local fileData = {}
	if string.sub(file, 1, 2) == ">>" then
		file = string.sub(file, 3, -1)
		lines = file:split("\n")
		for i = 1, #lines do
			fileData[#fileData+1] = lines[i]
		end
	else
		for filed in lfs.lines(CUR_GAME.."/scripts/"..file..".spt") do
			fileData[#fileData+1] = filed
		end
	end

	return processScriptFile(fileData)
end


local scriptSystem = { running = false }

function scriptSystem:init(arg)
	print("Initializing Scripting System")
	self.loadedFile = ""
	self.running = false
	self.scriptLocation = -1
	self.parent = arg.parent

	if arg.parentIsEntity == true then
		self.parentIsEntity = true
		self.thisEntity = arg.thisEntity
	end

	self.font = font.dialog
	self.lineSpacing = self.font:getHeight() + 1
	self._dbgDrawWidth = arg.dbgDrawWidth or 640
	self._dbgDrawHeight = arg.dbgDrawHeight or 480

	self.cron = require 'library.cron'
end

function scriptSystem:loadScript(file)
	if self.running then return false end
	self.data = loadScript(file)
	self.loadedFile = string.sub(file, 1, 2) == ">>" and "CUSTOM" or file..".spt"
end

function scriptSystem:run(callingEntity)
	self.callingEntity = callingEntity or ""
	self.queryAnswer = 0
	self.queryAnswerText = ""
	self.scriptLocation = 0
	self.commandFinished = true
	self.running = true
end

function scriptSystem:stop()
	self.callingEntity = nil
	self.scriptLocation = -1
	self.running = false
end

function scriptSystem:isRunning()
	return self.running
end

function scriptSystem:returnModuleContext(m)
	m = string.lower(m)
	if not scriptModule[m] then return {command = "UNDEFINED", editorBGColor = {0,0,0}, editorIcon = "end"} end
	return scriptModule[m]
end

function scriptSystem:getModuleList()
	return scriptModule
end

function scriptSystem:update(dt)
	if not self.running or self.scriptLocation < 0 then return false end
	self.cron.update(dt)

	if self.commandFinished then
		self:increment()
	end

	if self.scriptLocation > 0 then
		local cmd = string.lower(self.data.lines[self.scriptLocation].command)
		if scriptModule[cmd] then
			if scriptModule[cmd].update then
				scriptModule[cmd]:update(self, dt)
			end
		end
	end

end

function scriptSystem:increment()
	self.scriptLocation = self.scriptLocation + 1
	if self.scriptLocation > 1 then
		local cmd = string.lower(self.data.lines[self.scriptLocation-1].command)
		if scriptModule[cmd] then
			if scriptModule[cmd].close then
				scriptModule[cmd]:close(self)
			end
		end
	end
	if self.scriptLocation > #self.data.lines then
		self.scriptLocation = -1
		self.running = false
		return
	end
	self:runCommand(self.scriptLocation)
end

function scriptSystem:finishCommand()
	if not self.running then
		print("Script not running")
		return false
	end
	self.commandFinished = true
end

function scriptSystem:draw()
	if not self.running then return false end

	-- local x, y = self._dbgDrawWidth - (RENDERWIDTH + 20), self._dbgDrawHeight - (RENDERHEIGHT + 20)
	-- lgr.push()
	-- lgr.translate(x, y)
	-- lgr.setFont(font.dialogbold)
	-- lgr.setColor(0,0,0,200)
	-- lgr.print("Test Area:", 1, -9)
	-- lgr.setColor(1,1,1)
	-- lgr.print("Test Area:", 0, -10)
	-- lgr.setColor(0,0,0,100)
	-- lgr.rectangle("fill", -1+4, -1+4, RENDERWIDTH + 2, RENDERHEIGHT + 2)
	-- lgr.setColor(255,255,255,200)
	-- lgr.rectangle("fill", -1, -1, RENDERWIDTH + 2, RENDERHEIGHT + 2)
	-- lgr.setScissor(x, y, RENDERWIDTH, RENDERHEIGHT)
	-- lgr.setColor(100,0,160,255)
	-- lgr.rectangle("fill", 0, 0, RENDERWIDTH, RENDERHEIGHT)
	if self.scriptLocation > 0 then
		local cmd = string.lower(self.data.lines[self.scriptLocation].command)
		if scriptModule[cmd] then if scriptModule[cmd].draw then scriptModule[cmd]:draw(self) end end
	end
	-- lgr.setScissor()
	-- lgr.pop()
end

function scriptSystem:keypressed(k, r)
	if not self.running then return false end
	if self.scriptLocation > 0 then
		local cmd = string.lower(self.data.lines[self.scriptLocation].command)
		if scriptModule[cmd] then if scriptModule[cmd].keypressed then scriptModule[cmd]:keypressed(self, k, r) end end
	end
	return true
end

function scriptSystem:runCommand(line)
	if self.commandFinished == false then return end
	self.currentScriptLine = line
	local l = self.data.lines[line].command
	local a = self.data.lines[line].args[1]
	if line < #self.data.lines then
		self.nextCommand = self.data.lines[line+1].command
	else
		self.nextCommand = "ENDSCRIPT"
	end
	self.data.lines[line].completed = self.data.lines[line].completed + 1
	self.commandFinished = false
	if scriptModule[string.lower(l)] then
		scriptModule[string.lower(l)]:init(self, a, self.data.lines[line].dontWait)
	else
		self.commandFinished = true
	end
end





local function drawSpinningDesign(x, y, s)
	local lw = lgr.getLineWidth()
	lgr.setLineWidth(0.5)

	local sides = 3
	local fill = "fill"

	-- lgr.setColor(0,0,0)
	-- lgr.push()
	-- lgr.translate(x, y)
	-- lgr.scale(s)

	-- lgr.push()
	-- lgr.rotate(tmr.getTime() * 6)
	-- lgr.circle(fill, 0, 0, 7, sides)
	-- lgr.pop()

	-- lgr.push()
	-- lgr.rotate(tmr.getTime() * 2)
	-- lgr.circle(fill, 0, 0, 7, sides)
	-- lgr.pop()

	-- lgr.push()
	-- lgr.rotate(tmr.getTime() * 4)
	-- lgr.circle(fill, 0, 0, 7, sides)
	-- lgr.pop()

	-- lgr.pop()

	lgr.setBlendMode("add")

	local o = 255
	lgr.setColor(0,0,255,o)
	lgr.push()
	lgr.translate(x, y)
	-- lgr.print("Hello", -1, 0)
	lgr.scale(s)
	lgr.rotate(tmr.getTime() * 6)
	-- lgr.circle("fill", 0, 0, 4, sides)
	lgr.circle(fill, 0, 0, 5, sides)
	lgr.pop()

	lgr.setColor(0,255,0,o)
	lgr.push()
	lgr.translate(x, y)
	-- lgr.print("Hello", 0, 0)
	lgr.scale(s)
	lgr.rotate(tmr.getTime() * 2)
	-- lgr.circle("fill", 0, 0, 4, sides)
	lgr.circle(fill, 0, 0, 5, sides)
	lgr.pop()

	lgr.setColor(255,0,0,o)
	lgr.push()
	lgr.translate(x, y)
	-- lgr.print("Hello", 1, 0)
	lgr.scale(s)
	lgr.rotate(tmr.getTime() * 4)
	-- lgr.circle("fill", 0, 0, 4, sides)
	lgr.circle(fill, 0, 0, 5, sides)
	lgr.pop()

	lgr.setLineWidth(lw)
	lgr.setBlendMode("alpha")
end

messageBoxModule = {}

local colors = {
	dark = {
		background = {0,0,0,0.78},
		shadow = {0,0.29,0.78,0.55},
		normal = {1,1,1,1},
		red = {1,0,0,1},
		orange = {1,0.47,0,1},
		yellow = {1,1,0,1},
		green = {0,1,0,1},
		blue = {0.3,1,1,1},
		purple = {1,0,1,1},
	},
	light = {
		background = {1,1,1,0.9},
		shadow = {0,0,0,0.1},
		normal = {0,0,0,1},
		red = {0.78,0,0,1},
		orange = {0.78,0.29,0,1},
		yellow = {0.66,0.66,0,1},
		green = {0,0.78,0,1},
		blue = {0,0,0.78,1},
		purple = {0.78,0,0.78,1},
	},
}

function messageBoxModule:setup(v)
	local v = v or {}
	self.message = v.message or v.msg or v.text or "<<NO TEXT TO DISPLAY>>"
	self.typeSpeed = v.typeSpeed or v.typespeed or v.speed or 30
	self.width = 256 - 16
	self.height = 40
	self.x = v.x or (RENDERWIDTH - self.width) / 2
	self.y = v.y or RENDERHEIGHT - (self.height + 8)
	self.z = v.z or 1000000
	self.caret = 0
	self.font = font.dialog

	self.parent = v.parent or gameState.scriptSystem

	self.letterSpacing = v.letterSpacing or 1
	self.lineHeight = self.font:getLineHeight()
	self.fontHeight = self.font:getHeight()
	self.lineSpacing = v.lineSpacing or 0

	self.answers = v.answers or {"111", "222"}
	self.answer = 1
	self.isQuery = v.isQuery or false

	self.style = v.style or "dark"

	self.transition = 0
	self.transSpeed = 0.3
	self.transEase = "linear"

	self.transTween = flux.to(self, self.transSpeed, { transition = 1 }):ease(self.transEase):oncomplete(function()
		self:setupText()
	end)
	self.finished = false
	self.canDismiss = false
	self.spinnerScale = 0
end

function messageBoxModule:update(dt)
	if self.caret >= #self.message and not self.finished then
		self.finished = true
		self.parent.cron.after(0.2, function()
			self.canDismiss = true
		end)
	end

	if self.canDismiss and self.spinnerScale <= 0 then
		flux.to(self, 1, { spinnerScale = 1 }):ease("elasticout")
	end
end

function messageBoxModule:draw()
	lgr.push()
	lgr.translate(self.x, self.y)

	lgr.push()
	lgr.translate(self.width / 2, self.height / 2)
	lgr.scale(1, self.transition)
	lgr.translate(-(self.width / 2), -(self.height / 2))
	lgr.setColor(colors[self.style].background[1], colors[self.style].background[2], colors[self.style].background[3], colors[self.style].background[4] * self.transition)
	lgr.rectangle("fill", 0, 0, self.width, self.height, 4, 4, 8)
	lgr.pop()

	if self.letter then
		lgr.push()
		lgr.translate(8, 8)
		if #self.message > 0 then
			for i = 1, #self.message do
				local c = self.letter[i]
				if c then
					lgr.push()
					lgr.translate(c.x, c.y)
					lgr.scale(1, math.min(c.scale,1))
					lgr.rotate((1 - c.scale) / 2)
					lgr.setFont(self.letter[i].font)
					if c.shadow then
						lgr.setColor(c.color[1]/2, c.color[2]/2, c.color[3]/2, c.color[4])
						-- lgr.setColor(colors[self.style].shadow)
						lgr.print(c.char, c.shadowOffsetX, c.shadowOffsetY or 1)
					end
					-- lgr.setColor(255,0,0,100)
					-- lgr.rectangle("fill", 0, 0, c.w, 10)
					lgr.setColor(c.color)
					lgr.print(c.char, 0, 0)
					lgr.pop()
				end
			end
		end

		if self.isQuery and self.finished then
			local x = 0
			for i = 1, #self.answers do
				lgr.setColor(colors[self.style].normal)
				if self.answer == i then lgr.setColor(colors[self.style].yellow) end
				lgr.print(self.answers[i], x + 20, 24)
				x = x + (self.font:getWidth(self.answers[i]) + 40)
			end
		end
		lgr.pop()
	end

	if self.canDismiss and not self.isQuery then
		drawSpinningDesign(self.width - 15.5, 39.5, 1 * self.spinnerScale)
	end

	lgr.pop()
end

function messageBoxModule:setupText()
	self.letter = {}
	local lastSpace = 0
	local caretLocation = 0
	local lineTop = 0
	local spaces = {}
	local j = 1
	local color = colors[self.style].normal
	local shadow = false
	local d = 0
	self._curFont = self.font
	while not done do
		local skip = false
		local f = self.font
		local isSpace = false
		local sox = 0
		local soy = 0
		local adjust = 0
		local c = self.message:sub(j,j)
		if c == "\\" then
			c = self.message:sub(j,j+1)
		end
		if c == "{" then
			skip = true
			local cmd = self.message:sub(j+1,j+1)
			if cmd == "B" then
				self._curFont = font.dialogbold
				j = j + 2
				adjust = 1
			elseif cmd == "N" then
				self._curFont = font.dialog
				j = j + 2
				adjust = 1
			elseif cmd == "C" then
				local clr = self.message:sub(j+2,j+2)
				if clr == "R" then
					color = colors[self.style].red
				elseif clr == "O" then
					color = colors[self.style].orange
				elseif clr == "Y" then
					color = colors[self.style].yellow
				elseif clr == "G" then
					color = colors[self.style].green
				elseif clr == "B" then
					color = colors[self.style].blue
				elseif clr == "V" then
					color = colors[self.style].purple
				elseif clr == "D" then
					color = colors[self.style].normal
				end
				j = j + 3
				adjust = 1
			elseif cmd == "S" then
				shadow = not shadow
				j = j + 2
				adjust = 1
			elseif cmd == "W" then
				local loc = -1
				local loops = 0
				local n = j
				while loc == -1 do
					n = n + 1
					local ch = self.message:sub(n, n)
					if ch == "}" then break end
					loops = loops + 1
					if loops > 10 then break end
				end
				local time = tonumber(self.message:sub(j+2,n-1))
				d = d + time
				j = n
				adjust = 1
			end
		end
		local w = 0 - adjust
		if not skip then
			w = self._curFont:getWidth(c) - (1 + adjust)
		end
		if c == " " then
			lastSpace = caretLocation
			spaces[#spaces+1] = {
				char = j,
				location = lastSpace + w + self.letterSpacing
			}
		elseif c == "\\n" then
			lineTop = lineTop + self.lineHeight * self.fontHeight + self.lineSpacing
			caretLocation = -1
			j = j + 1
			w = 0
			d = d + 0.2
		elseif c == "\\t" then
			j = j + 1
			w = 23
		end
		if shadow then
			if self.style == "dark" then
				sox, soy = 0, -1
			else
				sox, soy = 0, 1
			end
		end
		if c ~= " " and c ~= "\\n" and c ~= "\\t" and not skip then
			-- print(c, w, caretLocation, lastSpace)
			local x = caretLocation
			local y = lineTop
			local rgb = colors[self.style].normal

			self.letter[j] = {
				char = c,
				width = w,
				color = {rgb[1],rgb[2],rgb[3],0.79},
				endColor = color,
				x = x,
				y = y + 8,
				w = w,
				endY = y,
				font = self._curFont,
				scale = 0,
				shadow = shadow,
				shadowOffsetX = sox,
				shadowOffsetY = soy,
			}
			self.letter[j].tween1 = flux.to(self.letter[j], 0.2, { scale = 1, y = self.letter[j].endY }):ease("backout"):delay(d)
			self.letter[j].tween2 = flux.to(self.letter[j].color, 0.1, { [1] = self.letter[j].endColor[1], [2] = self.letter[j].endColor[2], [3] = self.letter[j].endColor[3], [4] = self.letter[j].endColor[4] }):ease("linear"):delay(d+0.2)
			d = d + 0.025
		end
		caretLocation = caretLocation + w + self.letterSpacing
		j = j + 1
		if j > #self.message then
			self.parent.cron.after(d + 0.2, function()
				self.finished = true
				self.parent.cron.after(0.1, function()
					self.canDismiss = true
				end)
			end)
			break
		end
	end
end

function messageBoxModule:previousAnswer()
	self.answer = math.max(messageBoxModule.answer - 1, 1)
end

function messageBoxModule:nextAnswer()
	self.answer = math.min(messageBoxModule.answer + 1, #messageBoxModule.answers)
end

function messageBoxModule:getAnswer()
	return self.answer, self.answers[self.answer]
end

function messageBoxModule:speedUp(s)
	if not self.letter then return false end
	if self.transition < 1 then return false end
	for i = 1, #self.message do
		local l = self.letter[i]
		if l then
			if l.tween1 then l.tween1:stop() end
			if l.tween2 then l.tween2:stop() end
			l.tween1 = flux.to(l, 0.1, { scale = 1, y = l.endY }):ease("linear")
			l.tween2 = flux.to(l.color, 0.1, { [1] = l.endColor[1], [2] = l.endColor[2], [3] = l.endColor[3], [4] = l.endColor[4] }):ease("linear")
		end
	end

	self.parent.cron.after(0.15, function()
		self.finished = true
		self.parent.cron.after(0.1, function() self.canDismiss = true end)
	end)
end

function messageBoxModule:close(func)
	if self.transition < 1 then return false end
	for i = 1, #self.message do
		local l = self.letter[i]
		if l then
			if l.tween1 then l.tween1:stop() end
			if l.tween2 then l.tween2:stop() end
			l.tween1 = flux.to(l, 0.2, { y = l.endY + math.random(-20,-10) }):ease("linear")
			l.tween2 = flux.to(l.color, 0.2, { [4] = 0 }):ease("linear")
		end
	end

	self.transTween = flux.to(self, 0.2, { transition = 0, spinnerScale = 0 }):delay(0.01):oncomplete(function()
		if func then func() end
	end)
end




scriptSystem:init(arg)
return scriptSystem
