local lgr = love.graphics
local win = love.window
local kbd = love.keyboard
local lms = love.mouse

local sstack = {}

function sstack:init(...)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(...)
	return _i
end
function sstack:setup(...)
	local arg = ... or {}
	self.drawWidth = arg.drawWidth or WINDOW_WIDTH
	self.drawHeight = arg.drawHeight or WINDOW_HEIGHT
	self:resize(WINDOW_WIDTH, WINDOW_HEIGHT)
	self.useScissor = arg.useScissor or false
	self.stateStack = {}
	self.stateStackLevel = 0

	self.isSubStack = arg.isSubStack or false
	self.parentStack = arg.parentStack or nil

	self.useCanvas = arg.useCanvas or false

	self.screenFade = require "library.transition"
	self.screenFade:init { drawWidth = self.drawWidth, drawHeight = self.drawHeight }

	self.canvas = lgr.newCanvas(self.drawWidth, self.drawHeight, {dpiscale = 1})
	-- self.canvas:setFilter("linear", "nearest")

	self.useShader = false
end

function sstack:push(s, ...)
	if not s then return false end
	if s.loaded then return false end

	local n = #self.stateStack or 0
	if self.stateStack then
		if n > 0 then
			if self.stateStack[self.stateStackLevel].blur then
				self.stateStack[self.stateStackLevel]:blur()
			end
		end
	end
	self.stateStack[n + 1] = s
	self.stateStackLevel = #self.stateStack

	if s.canBeDuplicated then
		self.stateStack[self.stateStackLevel] = s:load(...)
		self.stateStack[self.stateStackLevel].parentStack = self
		self.stateStack[self.stateStackLevel].parentDrawWidth = self.drawWidth
		self.stateStack[self.stateStackLevel].parentDrawHeight = self.drawHeight
	else
		self.stateStack[self.stateStackLevel].parentStack = self
		self.stateStack[self.stateStackLevel].parentDrawWidth = self.drawWidth
		self.stateStack[self.stateStackLevel].parentDrawHeight = self.drawHeight
		self.stateStack[self.stateStackLevel]:load(...)
	end
	self.stateStack[self.stateStackLevel].id = #self.stateStack

	self.stateStack[self.stateStackLevel].stackLevel = #self.stateStack

	if self.stateStack[self.stateStackLevel].focus then self.stateStack[self.stateStackLevel]:focus() end
end

function sstack:back(...)
	if #self.stateStack == 1 then print("This is the top level of the stack. You cannot go higher. This is not an error.") return false end
	if self.stateStack[self.stateStackLevel].blur then self.stateStack[self.stateStackLevel]:blur() end
	self.stateStack[self.stateStackLevel]:unload()
	self.stateStack[#self.stateStack] = nil
	self.stateStackLevel = #self.stateStack
	if self.stateStack[self.stateStackLevel].focus then self.stateStack[self.stateStackLevel]:focus(nil, ...) end
end

function sstack:replace(s, ...)
	local arg = ...
	if not s then return false end
	if s.loaded then return false end

	self:release()
	self.stateStack = {}

	self:push(s, arg)
	self.ignoreDrawForNextFrame = true
end

function sstack:release()
	for i = self.stateStackLevel, 1, -1 do
		if self.stateStack[i].blur then self.stateStack[i]:blur() end
		self.stateStack[i]:unload()
		self.stateStack[i] = nil
	end
end

function sstack:getTopContext()
	return self.stateStack[self.stateStackLevel]
end

function sstack:changeDrawDimensions(w, h)
	if not w or not h then return false end
	-- print(w, h)
	self.drawWidth = w
	self.drawHeight = h
	self.stateStack[self.stateStackLevel].parentDrawWidth = self.drawWidth
	self.stateStack[self.stateStackLevel].parentDrawHeight = self.drawHeight
	self.canvas = lgr.newCanvas(self.drawWidth, self.drawHeight, {dpiscale = 1})
	self:resize(WINDOW_WIDTH, WINDOW_HEIGHT)
	self.screenFade:init { drawWidth = self.drawWidth, drawHeight = self.drawHeight }
end

function sstack:goto(level, ...)
	if not level then return false end
	if level >= self.stateStackLevel then return false end
	for i = self.stateStackLevel, level + 1, -1 do
		self.stateStack[i]:blur()
		self.stateStack[i]:unload()
		self.stateStack[i] = nil
	end

	self.stateStackLevel = #self.stateStack
	self.stateStack[self.stateStackLevel]:focus(nil, ...)
end


local min = math.min
local max = math.max
local floor = math.floor
-- Standard Callbacks
function sstack:update(dt)
	-- self.debugCanvas:renderTo(function()
	-- 	lgr.clear(0,0,0,0)
	-- end)

	local ox, oy, s = self.drawHorizOffset, performanceGraph.enabled and 0 or self.drawVertOffset, self.drawScale
	if self.isSubStack then
		ox, oy, s = self.parentStack.drawHorizOffset, self.parentStack.drawVertOffset, self.parentStack.drawScale
	end
	self.mouseX = math_clamp(floor((lms.getX() - ox) / (s)), 0, self.drawWidth - 1)
	self.mouseY = math_clamp(floor((lms.getY() - oy) / (s)), 0, self.drawHeight - 1)

	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if self.stateStack[i].update then
			if i == #self.stateStack or self.stateStack[i].backgroundUpdate then
				self.stateStack[i]:update(dt)
			end
		end
	end
	self.screenFade:update(dt)
end

local function _drawNoStackPlaceholder(self)
	lgr.setColor(1,0,0)
	lgr.circle("fill", 0, 0, 20)
	lgr.setColor(1,1,0)
	lgr.circle("fill", self.drawWidth, 0, 20)
	lgr.setColor(0,1,0)
	lgr.circle("fill", self.drawWidth, self.drawHeight, 20)
	lgr.setColor(0,0,1)
	lgr.circle("fill", 0, self.drawHeight, 20)

	lgr.setColor(1,1,1)
	lgr.setFont(font.dialog)
	lgr.printf("This is the State Stack drawing area!\nCurrently there are no states pushed to the stack!\nThe dimensions of this State Stack are " .. self.drawWidth .. " x " .. self.drawHeight .. ".\nThe current draw scale is " .. self.drawScale .. ".", 0, 10, self.drawWidth, "center")
	lgr.rectangle("line", 0.5, 0.5, self.drawWidth - 1, self.drawHeight - 1)
	lgr.line(0, 0, self.drawWidth, self.drawHeight)
	lgr.line(self.drawWidth, 0, 0, self.drawHeight)
end
function sstack:draw()
	-- if self.useScissor then lgr.setScissor(self.drawHorizOffset, self.drawVertOffset, self.drawWidth * self.drawScale, self.drawHeight * self.drawScale) end

	self.canvas:renderTo(function()
		lgr.clear(0,0,0,0)
		lgr.push()
		lgr.origin()

		if not self.ignoreDrawForNextFrame then
			if #self.stateStack == 0 then
				_drawNoStackPlaceholder(self)
			else
				if not self.stateStack[self.stateStackLevel] then return false end
				for i = 1, self.stateStackLevel do
					if self.stateStack[i].draw then
						if i == #self.stateStack or self.stateStack[i].backgroundDraw then
							self.stateStack[i]:draw(dt)
						end
					end
				end
			end
		end

		lgr.pop()
		self.screenFade:draw()
	end)

	lgr.push()
	if not self.isSubStack then
		lgr.translate(self.drawHorizOffset, performanceGraph.enabled and 0 or self.drawVertOffset)
		lgr.scale(self.drawScale)
	end

	lgr.setColor(1,1,1)
	-- if self.useShader then
	-- 	lgr.setShader(crtShader)
	-- 	crtShader:send("inputSize", {self.drawWidth/2, self.drawHeight/2})
	-- 	crtShader:send("outputSize", {self.drawWidth, self.drawHeight})
	-- 	crtShader:send("textureSize", {self.drawWidth, self.drawHeight})
	-- end
	lgr.draw(self.canvas, 0, 0)
	-- if self.useShader then
	--     lgr.setShader()
	-- end

	lgr.pop()

	self.ignoreDrawForNextFrame = false

	-- lgr.setColor(1,1,1)
	-- lgr.draw(self.debugCanvas, 0, (self.drawHeight * self.drawScale), 0, 2 * win.getDPIScale())

	-- if self.useScissor then lgr.setScissor() end

	-- lgr.push()
	-- if not self.isSubStack then
	-- 	lgr.scale(2 * win.getDPIScale())
	-- end
	-- lgr.setColor(1,1,1)
	-- lgr.setFont(font.dialog)
	-- lgr.print(#self.stateStack, 0, 0)
	-- for i = 1, #self.stateStack do
	-- 	if self.stateStack[i].name then
	-- 		lgr.print(i .. ": " .. self.stateStack[i].name or "What", 0, i * 12)
	-- 	else
	-- 		lgr.print(i .. ": Unnamed", 0, i * 12)
	-- 	end
	-- end
	-- lgr.pop()
end

-- Keyboard
function sstack:keypressed(k, r)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if self.stateStack[i].keypressed then
			if i == #self.stateStack or self.stateStack[i].backgroundKeyEvents then
				self.stateStack[i]:keypressed(k, u)
			end
		end
	end
end

function sstack:keyreleased(k, r)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if self.stateStack[i].keyreleased then
			if i == #self.stateStack or self.stateStack[i].backgroundKeyEvents then
				self.stateStack[i]:keyreleased(k, u)
			end
		end
	end
end

function sstack:textinput(k, u)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if self.stateStack[i].textinput then
			if i == #self.stateStack then
				self.stateStack[i]:textinput(k, u)
			end
		end
	end
end

-- Mouse
function sstack:mousefocus(f)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].mousefocus then
			self.stateStack[i]:mousefocus(f)
		end
	end
end

function sstack:wheelmoved(x, y)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].wheelmoved then
			self.stateStack[i]:wheelmoved(x, y)
		end
	end
end

function sstack:mousemoved(x, y, dx, dy)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].mousemoved then
			self.stateStack[i]:mousemoved(self.mouseX, self.mouseY, dx / self.drawScale, dy / self.drawScale)
		end
	end
end

function sstack:mousepressed(x, y, button)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].mousepressed then
			self.stateStack[i]:mousepressed(self.mouseX, self.mouseY, button)
		end
	end
end

function sstack:mousereleased(x, y, button)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].mousereleased then
			self.stateStack[i]:mousereleased(self.mouseX, self.mouseY, button)
		end
	end
end

function sstack:doubleclick(x, y, button)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].doubleclick then
			self.stateStack[i]:doubleclick(self.mouseX, self.mouseY, button)
		end
	end
end

function sstack:getMousePixelPosition()
	-- local x, y = lms.getX(), lms.getY()
	-- x = x - self.drawHorizOffset
	-- y = y - self.drawVertOffset
	-- print(self.drawHorizOffset, self.drawVertOffset)
	return self.mouseX, self.mouseY
end

function sstack:getMouseX()
	return self.mouseX
end

function sstack:getMouseY()
	return self.mouseY
end


-- GamePad / JoyStick
function sstack:gamepadaxis(joystick, axis, value)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].gamepadaxis then
			self.stateStack[i]:gamepadaxis(joystick, axis, value)
		end
	end
end

function sstack:gamepadpressed(joystick, button)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].gamepadpressed then
			self.stateStack[i]:gamepadpressed(joystick, button)
		end
	end
end

function sstack:gamepadreleased(joystick, button)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].gamepadreleased then
			self.stateStack[i]:gamepadreleased(joystick, button)
		end
	end
end

function sstack:joystickadded(joystick)

end

function sstack:joystickaxis(joystick, axis, value)

end

function sstack:joystickhat(joystick, hat, direction)

end

function sstack:joystickpressed(joystick, button)

end

function sstack:joystickreleased(joystick, button)

end

function sstack:joystickremoved(joystick)

end

-- Miscellaneous
function sstack:focus(f, arg)
	if not self.stateStack[self.stateStackLevel] then return false end
	for i = 1, self.stateStackLevel do
		if i == #self.stateStack and self.stateStack[i].focus then
			self.stateStack[i]:focus(f, arg)
		end
	end
end

function sstack:visible(f)

end

function sstack:resize(w, h)
	print("State Stack resized to:", w, h)

	-- Calculate the scaling and draw offset
	local sw, sh = w / self.drawWidth, h / self.drawHeight

	local aspect_window = w / h
	local aspect_game = self.drawWidth / self.drawHeight

	if aspect_window >= aspect_game then
		self.drawScale = sh
	else
		self.drawScale = sw
	end

	local _hSpace = w - (self.drawWidth * self.drawScale)
	local _vSpace = h - (self.drawHeight * self.drawScale)
	self.drawHorizOffset = _hSpace / 2
	self.drawVertOffset = _vSpace / 2

	if self.stateStack then
		for i = 1, #self.stateStack do
			self.stateStack[i].parentDrawWidth = math.floor(self.drawWidth)
			self.stateStack[i].parentDrawHeight = math.floor(self.drawHeight)
			self.stateStack[i]:resize(math.floor(self.drawWidth), math.floor(self.drawHeight))
		end
	end

	-- if _vSpace > 0 then
	-- 	local cw, ch = w / 2 / win.getDPIScale(), (_vSpace / 4)
	-- 	print("Debug Canvas Size:", cw, ch)
	-- 	self.debugCanvas = lgr.newCanvas(cw, ch, {dpiscale = 1})
	-- 	self.debugCanvas:renderTo(function()
	-- 		lgr.setColor(1,1,1)
	-- 		for i = 0, cw, 2 do
	-- 			lgr.line(i+0.5, 0, i+0.5, ch)
	-- 		end
	-- 		for i = 0, ch, 2 do
	-- 			lgr.line(0, i+0.5, cw, i+0.5)
	-- 		end
	-- 	end)
	-- end
end

return sstack
