local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local lms = love.mouse
local tmr = love.timer

local SHOWDEBUG = false

local toolbarButton = {}

function toolbarButton:create(arg)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(arg or {})
	return _i
end
function toolbarButton:setup(arg)
	self.parent = arg.parent
	self.nameID = arg.nameID
	self.vertical = self.parent.vertical

	self.image = arg.image or nil
	self.quad = arg.quad or nil
	self.drawFunction = arg.drawFunction or nil
	self.functionVars = arg.functionVars or nil
	self.canvas = lgr.newCanvas(17,17, {dpiscale = 1})

	self.drawOX = arg.drawOX or 0
	self.drawOY = arg.drawOY or 0

	local vx, vy, vw, vh = 0, 0, 16, 16
	if self.quad then
		vx, vy, vw, vh = self.quad:getViewport()
	end

	self.label = arg.label or nil
	self.altLabel = arg.altLabel or nil
	self.showLabel = arg.showLabel or false

	self.x = 0
	self.y = 0
	if self.vertical then
		self.w = self.parent.w
		self.h = self.parent.buttonHeight
		self.showLabel = true
	else
		self.w = self.parent.hPadding + vw + self.parent.hPadding
		if self.showLabel then
			self.w = self.w + (self.altLabelWidth or 48)
		end
		self.h = self.parent.buttonHeight
	end

	self.isPressed = false

	self.disabled = arg.disabled

	self.font = self.parent.font

	if self.label then
		self._labelWidth = self.font:getWidth(self.label)
		self._labelHeight = self.font:getHeight()
	end
	self._imageOffsetY = math.ceil((self.h - vh) / 2)
	self._labelOffsetY = self.parent.vPadding
	if self.altLabel then
		local calc = vw + self.parent.hPadding + self.font:getWidth(self.altLabel)
		-- self._imageOffsetX = (self.w - (self.parent.hPadding * 2) - calc) / 2
		self._labelOffsetX = vw + self.parent.hPadding + 5
	else
		-- self._imageOffsetX = self.parent.hPadding
	end

	self.highlightColor = arg.highlightColor or self.parent.color.highlight

	self.callback = arg.callback or nil
	self.click = arg.click or nil
	self.cbmousepressed = arg.mousepressed or nil
	self.cbmousereleased = arg.mousereleased or nil

	self.id = arg.id or 1
end

function toolbarButton:update(dt)
	if self.callback then self.callback(self, dt) end
end

function toolbarButton:draw()
	lgr.setColor(self.parent.color.background1)
	if self.hovering then lgr.setColor(self.parent.color.background2) end
	if self.isPressed then lgr.setColor(self.parent.color.selection) end
	if self.isSelected then lgr.setColor(self.parent.color.background1) end
	if self.isHighlighted then lgr.setColor(self.highlightColor) end
	lgr.rectangle("fill", self.x, self.y, self.w, self.h)

	self.canvas:renderTo(function()
		lgr.clear(DEBUGMODE and 1 or 0,0,0, DEBUGMODE and 1 or 0)

		if self.drawFunction then
			local vars = self.functionVars or {}
			vars.toolbarMode = true
			if self.parent.iconShadow then
				lgr.setColor(0,0,0,0.39)
				lgr.push()
				lgr.translate(self.parent.shadowOffset, self.parent.shadowOffset)
				self.drawFunction(vars.entity, vars)
				lgr.pop()
			end
			lgr.setColor(1,1,1)
			vars.x = 0
			vars.y = 0
			self.drawFunction(vars.entity, vars)
		elseif self.image then
			if self.parent.iconShadow then
				lgr.setColor(0,0,0,0.39)
				lgr.draw(self.image, self.quad, self.parent.shadowOffset, self.parent.shadowOffset)
			end
			lgr.setColor(self.parent.color.text)
			if self.disabled then
				lgr.setColor(1,1,1,0.39)
			end
			lgr.draw(self.image, self.quad, 0, 0)
		end

	end)

	lgr.push()
	lgr.translate(math.floor(self.x + self.parent.hPadding), math.floor(self.y + self._imageOffsetY))
	-- lgr.setColor(0,0,0)
	-- lgr.draw(self.canvas, self.drawOX or 0 + 1, self.drawOY or 0 + 1)
	lgr.setColor(1,1,1)
	lgr.draw(self.canvas, self.drawOX or 0, self.drawOY or 0)
	lgr.pop()

	if self.altLabel and self.showLabel then
		lgr.push()
		lgr.translate(self.x + self._labelOffsetX, self.y + self._labelOffsetY)
		-- if self.hasSubText then
		-- 	y = 2
		-- 	lgr.setFont(font.tiny)
		-- 	lgr.printf(self.subText, 4, 12, self.w - 8, "center")
		-- end
		lgr.setFont(font.dialog)
		if self.parent.iconShadow then
			lgr.setColor(0,0,0,0.39)
			lgr.printf(self.altLabel, 1, 1, self.w - 8, "left")
		end
		lgr.setColor(1,1,1)
		if self.disabled then
			lgr.setColor(1,1,1,0.39)
		end
		lgr.printf(self.altLabel, 0, 0, self.w - 8, "left")
		lgr.pop()
	end
end

function toolbarButton:mousepressed(x, y, button)
	if self.cbmousepressed and not self.isPressed then self:cbmousepressed(x, y, button) end
	self.isPressed = true
end

function toolbarButton:mousereleased(x, y, button)
	if self.cbmousereleased and self.isPressed then self:cbmousereleased(x, y, button) end
	self.isPressed = false
end

function toolbarButton:mouseclick(x, y, button)
	if self.disabled then return false end
	if button == 1 then
		if self.click then self.click(self) end
	end
end



local toolbarSeparator = {}

function toolbarSeparator:create(arg)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(arg or {})
	return _i
end
function toolbarSeparator:setup(arg)
	self.parent = arg.parent
	self.nameID = arg.nameID
	self.vertical = self.parent.vertical

	self.x = 0
	self.y = 0
	if self.vertical then
		self.w = self.parent.w
		self.h = arg.w or 7
	else
		self.w = arg.w or 7
		self.h = self.parent.h
	end
	self.space = arg.space or false

	self.changeColor = arg.changeColor or nil

	self.hasBorder = self.parent.hasBorder
	self.borderColor = self.parent.borderColor

	self.callback = arg.callback or nil
	self.click = arg.click or nil

	self.id = arg.id or 1
end

function toolbarSeparator:update(dt) end

function toolbarSeparator:draw()
	lgr.push()
	lgr.translate(self.x, self.y)
	if not self.space then
		lgr.setColor(self.parent.color.background1)
		lgr.rectangle("fill", 0, 0, math.floor(self.w), self.h)
	else
		lgr.setColor(self.parent.color.background1)
		if self.vertical then
			lgr.rectangle("fill", 0, 0, self.w, math.floor(self.h / 2))
			lgr.rectangle("fill", 0, math.floor(self.h / 2) + 1, self.w, math.ceil(self.h / 2) - 1)

			if self.hasBorder then
				lgr.setColor(self.borderColor)
				lgr.setLineWidth(1)
				lgr.line(0, math.floor(self.h / 2) + 1, self.w, math.floor(self.h / 2) + 1)
			end
		else
			lgr.rectangle("fill", 0, 0, math.floor(self.w / 2), self.h)
			lgr.rectangle("fill", math.floor(self.w / 2) + 1, 0, math.ceil(self.w / 2) - 1, self.h)

			if self.hasBorder then
				lgr.setColor(self.borderColor)
				lgr.setLineWidth(1)
				lgr.line(math.floor(self.w / 2) + 1, 0, math.floor(self.w / 2) + 1, self.h)
			end
		end
	end
	lgr.pop()
end

function toolbarSeparator:mousepressed(x, y, button) end

function toolbarSeparator:mousereleased(x, y, button) end

function toolbarSeparator:mouseclick(x, y, button)
	if button == 1 then
		if self.click then self.click(self) end
	end
end



local toolbarLabel = {}

function toolbarLabel:create(arg)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(arg or {})
	return _i
end
function toolbarLabel:setup(arg)
	self.parent = arg.parent
	self.nameID = arg.nameID
	self.vertical = self.parent.vertical

	self.x = 0
	self.y = 0
	if self.vertical then
		self.w = self.parent.w
	else
		self.w = arg.w or 7
	end
	self.h = self.parent.buttonHeight

	self.expand = arg.expand or false

	self.textColor = arg.textColor or self.parent.color.text

	self.font = self.parent.font
	self.smallFont = self.parent.smallFont

	self.text = arg.text or nil
	self.subText = arg.subText or nil
	self.hasSubText = self.hasSubText or self.subText ~= nil or false

	self.callback = arg.callback or nil
	self.click = arg.click or nil

	self.id = arg.id or 1
end

function toolbarLabel:update(dt) end

function toolbarLabel:draw()
	lgr.push()
	lgr.translate(self.x, self.y)
	lgr.setColor(self.parent.color.background1)
	lgr.rectangle("fill", 0, 0, self.w, self.h)
	if self.text then
		lgr.setColor(self.textColor)
		local y = 7
		if self.hasSubText then
			y = 3
			lgr.setFont(self.smallFont)
			lgr.printf(self.subText, 4, 13, self.w - 8, "center")
		end
		lgr.setFont(self.font)
		lgr.printf(self.text, 4, y, self.w - 8, "center")
	end

	lgr.pop()
end

function toolbarLabel:mousepressed(x, y, button) end

function toolbarLabel:mousereleased(x, y, button) end

function toolbarLabel:mouseclick(x, y, button)
	if button == 1 then
		if self.click then self.click(self) end
	end
end


toolbarGUI = {}

function toolbarGUI:create(arg)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(arg or {})
	return _i
end
function toolbarGUI:setup(arg)
	local arg = arg or {}
	self.x = math.floor(arg.x) or 0
	self.y = arg.y or 0
	self.w = math.floor(arg.w) or 300
	self.h = arg.h or 19
	self.buttonHeight = self.h

	self.vertical = arg.vertical and true or false

	self.color = {
		background1 = arg.backgroundColor1 or {0,0,0,0.78},
		background2 = arg.backgroundColor2 or {1,0.6,0,0.9},
		selection = arg.selectionColor or {1,0,0,0.39},
		highlight = arg.highlightColor or {0.26,0.86,0.95,0.78},
		text = arg.textColor or {1,1,1},
		tooltipBackground = arg.tooltipBackgroundColor or {0,0,0,0.78},
		tooltipText = arg.tooltipTextColor or {1,1,1},
	}

	self.hasBorder = arg.borderColor and true or false
	self.borderColor = arg.borderColor or {0,0,0,0}

	self.font = arg.font or font.dialog
	self.smallFont = arg.smallFont or font.tiny

	self.hPadding = arg.hPadding or 5
	self.vPadding = arg.vPadding or 3

	self.iconShadow = true
	if arg.iconShadow == false then self.iconShadow = false end
	self.shadowOffset = arg.shadowOffset or 1

	self.tooltipAlignment = arg.tooltipAlignment or "bottom"

	self.parent = arg.parent

	self.children = {}
	self.childrenID = {}
end

function toolbarGUI:update(dt)
	local x, y = stateStack.mouseX, stateStack.mouseY
	self.mouseOver = pointInRectangle(x, y, self.x, self.y, self.w, self.h)

	if self.mouseOver then
		lms.setCursor(appCursors.normal)
	end

	self.hoverButton = -1
	for i = 1, #self.childrenID do
		local c = self.children[self.childrenID[i]]
		c:update(dt)
		if self.parent.focused then
			if pointInRectangle(x, y, c.x, c.y, c.w-1, c.h-1) then
				c.hovering = true
				self.hoverButton = self.childrenID[i]
			else
				c.hovering = false
			end
		end
	end
end

function toolbarGUI:draw()
	for i = 1, #self.childrenID do
		local tbc = self.children[self.childrenID[i]]
		tbc:draw()
		if SHOWDEBUG then
			local s = 3
			lgr.setColor(1,1,1)
			lgr.rectangle("fill", tbc.x, tbc.y, 1, s)
			lgr.rectangle("fill", tbc.x, tbc.y, s, 1)
			lgr.rectangle("fill", tbc.x + tbc.w - 1, tbc.y + tbc.h - s, 1, s)
			lgr.rectangle("fill", tbc.x + tbc.w - s, tbc.y + tbc.h - 1, s, 1)
		end
	end

	if self.hasBorder then
		lgr.setColor(self.borderColor)
		lgr.rectangle("line", self.x, self.y, self.w + 1, self.h + 1)
	end

	if self.vertical then

	else
		if self.hoverButton ~= -1 and self.hoverButton and self.parent.focused then
			if self.children[self.hoverButton].label then
				local x
				local y
				if self.tooltipAlignment == "top" then
					x = math.floor(self.children[self.hoverButton].x + (self.children[self.hoverButton].w / 2) - self.children[self.hoverButton]._labelWidth / 2)
					y = math.floor(self.y - (self.children[self.hoverButton]._labelHeight + 3))
				elseif self.tooltipAlignment == "bottom" then
					x = math.floor(self.children[self.hoverButton].x + (self.children[self.hoverButton].w / 2) - self.children[self.hoverButton]._labelWidth / 2)
					y = math.floor(self.y + self.h + 4)
				elseif self.tooltipAlignment == "left" then
					x = self.children[self.hoverButton].x - (self.children[self.hoverButton]._labelWidth + 3)
					y = self.y + 2
				end
				lgr.push()
				lgr.setFont(self.children[self.hoverButton].font)
				lgr.translate(math_clamp(x, 4, self.parent.parentDrawWidth - (self.children[self.hoverButton]._labelWidth + 4)), y)
				lgr.setColor(self.color.tooltipBackground)
				lgr.rectangle("fill", -3, -2, self.children[self.hoverButton]._labelWidth + 6, self.children[self.hoverButton]._labelHeight + 4, 2)
				lgr.setColor(self.color.tooltipText)
				lgr.printf(self.children[self.hoverButton].label, 0, 0, self.children[self.hoverButton]._labelWidth, "center")
				lgr.pop()
			end
		elseif not self.hoverButton then
			-- print("Hmm")
		end
	end
end

function toolbarGUI:mousepressed(x, y, button)
	if pointInRectangle(x, y, self.x, self.y, self.w, self.h) then
		for i = 1, #self.childrenID do
			local c = self.children[self.childrenID[i]]
			if pointInRectangle(x, y, c.x, c.y, c.w, c.h) then
				self.pressedButtonID = i
				c:mousepressed(x, y, button)
				-- print("ID", i)
				return true
			end
		end
		return true
	end
	return false
end

function toolbarGUI:mousereleased(x, y, button)
	for i = 1, #self.childrenID do
		local c = self.children[self.childrenID[i]]
		c:mousereleased(x, y, button)
		if pointInRectangle(x, y, c.x, c.y, c.w, c.h) and i == self.pressedButtonID then
			c:mouseclick(x, y, button)
			return true
		end
	end
	self.pressedButtonID = -1
	return false
end

function toolbarGUI:addButton(arg)
	arg.parent = self
	arg.nameID = arg.nameID or "tb_"..(#self.childrenID+1).."_"..math.random(1000,9999).."_"..math.random(1000,9999)
	self.childrenID[#self.childrenID+1] = arg.nameID
	self.children[arg.nameID] = toolbarButton:create(arg)
	self:reorderChildren()
end

function toolbarGUI:addSeparator(arg)
	arg.parent = self
	arg.nameID = arg.nameID or "tb_"..(#self.childrenID+1).."_"..math.random(1000,9999).."_"..math.random(1000,9999)
	self.childrenID[#self.childrenID+1] = arg.nameID
	self.children[arg.nameID] = toolbarSeparator:create(arg)
	self:reorderChildren()
end

function toolbarGUI:addLabel(arg)
	arg.parent = self
	arg.nameID = arg.nameID or "tb_"..(#self.childrenID+1).."_"..math.random(1000,9999).."_"..math.random(1000,9999)
	self.childrenID[#self.childrenID+1] = arg.nameID
	self.children[arg.nameID] = toolbarLabel:create(arg)
	self:reorderChildren()
end

function toolbarGUI:reorderChildren()
	local hasExpander = 0
	if not self.vertical then
		local remainingWidth = 0
		for i = 1, #self.childrenID do
			if self.children[self.childrenID[i]].expand then
				hasExpander = hasExpander + 1
			else
				remainingWidth = remainingWidth + self.children[self.childrenID[i]].w
			end
		end
		local expanderWidths = (self.w - remainingWidth) / hasExpander

		for i = 1, #self.childrenID do
			if self.children[self.childrenID[i]].expand then
				self.children[self.childrenID[i]].w = expanderWidths
			end
		end

		table.sort(self.children, function(a, b) return a.id < b.id end)
		local x = self.x
		for i = 1, #self.childrenID do
			local c = self.children[self.childrenID[i]]

			c.x = x
			c.y = self.y

			x = x + c.w
		end
	else
		table.sort(self.children, function(a, b) return a.id < b.id end)
		local y = self.y
		for i = 1, #self.childrenID do
			local c = self.children[self.childrenID[i]]

			c.x = self.x
			c.y = y

			y = y + c.h
		end
		self.h = y
	end
end

function toolbarGUI:isMouseOver()
	return self.mouseOver or (self.pressedButtonID or -1) > -1
end

function toolbarGUI:setLabelText(label, text)

end


function toolbarGUI:setWidth(w)
	self.w = w or 100
	self:reorderChildren()
end

function toolbarGUI:setPosition(x, y)
	self.x = x
	self.y = y
	self:reorderChildren()
end
