local lgr = love.graphics
local win = love.window
local kbd = love.keyboard
local tmr = love.timer
local lfs = love.filesystem

local timeSinceLastMouserelease = -1
local lastMouseClickX = -1
local lastMouseClickY = -1

local codes = {reset = "\x1B[m", red = "\x1B[91m", green = "\x1B[92m", yellow = "\x1B[93m", blue = "\x1B[94m", magenta = "\x1B[95m", cyan = "\x1B[96m"}
function cprint(colour, ...)
  if not codes[colour] then error("Undefined colour: " .. colour) end
  io.write(codes[colour])
  print(...)
  io.write(codes.reset)
end

-- aberration = lgr.newShader("library/shaders/aberration.frag")
fogShader = lgr.newShader("library/shaders/fogshader.frag")
-- waterShader = lgr.newShader("library/shaders/water.frag")
-- crtShader = lgr.newShader("library/shaders/CRT-Simple.frag")
outlineShader = lgr.newShader([[
	extern vec2 pixelsize;
	extern float size = 1;
	extern float smoothness = 1;

	vec4 effect(vec4 color, Image texture, vec2 uv, vec2 fc) {
		float a = 0;
		for(float y = -size; y <= size; ++y) {
			for(float x = -size; x <= size; ++x) {
				a += Texel(texture, uv + vec2(x * pixelsize.x, y * pixelsize.y)).a;
			}
		}
		a = color.a * min(1, a / (2 * size * smoothness + 1));

		return vec4(color.rgb, a);
	}
]])

nesShader = love.graphics.newShader[[
	extern vec4 color1;
	extern vec4 color2;
	extern vec4 color3;
	extern vec4 color4;

	vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords ){
		vec4 pixel = Texel(texture, texture_coords );//This is the current pixel color
		if (pixel.a == 0.0) { return vec4(0.0,0.0,0.0,0.0); }
		if (pixel.r < 0.25) { return vec4(color4.r,color4.g,color4.b,color4.a); }
		else if (pixel.r < 0.5) { return vec4(color3.r,color3.g,color3.b,color3.a); }
		else if (pixel.r < 0.75) { return vec4(color2.r,color2.g,color2.b,color2.a); }
		else { return vec4(color1.r,color1.g,color1.b,color1.a); }
		return pixel;
	}
]]
-- CUR_GAME = "game_adventure"
-- CUR_GAME = "game_proto"
-- CUR_GAME = "game_3d"
CUR_GAME = "game_window"
LAST_MAP_LOADED_IN_EDITOR = ""
USEPALETTE = true
THREEDEERESOLUTION = 1

function load_image(file, name)
	return lgr.newImage(love.data.decode("data", "base64", file))
	-- return lgr.newImage(love.filesystem.newFileData(file, name:gsub("_", "."), "base64"))
end

DEBUGPRINT = false
function __dprint(...)
	if not DEBUGPRINT then return false end
	print(...)
end

function getResolutionList()
	local _resolutionOptions = {
		{ w = 240, h = 160, nickname = "GBA" },
		{ w = 256, h = 144, nickname = "Widescreen GB" },
		{ w = 256, h = 240, nickname = "NES/SNES" },
		{ w = 320, h = 224, nickname = "Genesis" },
		{ w = 128, h = 128, nickname = "PICO-8" },
		{ w = 256, h = 256, nickname = "PICO-8 x 2" },
		{ w = 192, h = 128, nickname = "Wide PICO-8" },
		{ w = 400, h = 224, nickname = "Shovel Knight" },
		{ w = 272, h = 176, nickname = "BoI:Rebirth" },
		{ w = 480/2, h = 272/2, nickname = "1/2 Pocket Chip / PSP" },
		{ w = 480, h = 272, nickname = "Pocket Chip / PSP" },
		{ w = 160, h = 144, nickname = "Game Boy" },
		{ w = 848/2, h = 480/2, nickname = "1/2 480p" },
		{ w = 852/2, h = 480/2, nickname = "1/2 480p B" },
		{ w = 1280/2, h = 720/2, nickname = "1/2 720p / 1/3 1080p" },
		{ w = 1280/4, h = 720/4, nickname = "1/4 720p" },
		{ w = 1280/4, h = 11*16, nickname = "(Adjusted) 1/4 720p" },
		{ w = 1280/4, h = 800/4, nickname = "1/4 800p / DOS CGA" },
		{ w = 1920/2, h = 1080/2, nickname = "1/2 1080p" },
		{ w = 1920/4, h = 1080/4, nickname = "1/4 1080p" },
		{ w = 1920/2, h = 1200/2, nickname = "1/2 1200p" },
		{ w = 1920/3, h = 1200/3, nickname = "1/3 1200p" },
		{ w = 1920/4, h = 1200/4, nickname = "1/4 1200p" },
		{ w = 512, h = 288, nickname = "Tiny Heist" },
		{ w = 512, h = 384, nickname = "Macintosh Classic" },
		{ w = 512/2, h = 384/2, nickname = "1/2 Macintosh Classic" },
		{ w = 342, h = 192, nickname = "" },
		{ w = 1920, h = 1080, nickname = "1080p" },
		{ w = 1280, h = 720, nickname = "720p" },
	}
	table.sort(_resolutionOptions, function(a, b)
		if a.w * a.h ~= b.w * b.h then return a.w * a.h < b.w * b.h end
		if a.w / a.h ~= b.w / b.h then return a.w / a.h < b.w / b.h end
		return a.h < b.h
	end)

	for i = 1, #_resolutionOptions do
		local o = _resolutionOptions[i]
		local aspect = tostring(math_round(o.w / o.h, 3))
		if aspect == "1.778" then aspect = "16:9" end
		if string.sub(aspect, 1, 3) == "1.7" then aspect = "16:9*" end
		if aspect == "1.6" then aspect = "16:10" end
		if aspect == "1" then aspect = "Square" end
		_resolutionOptions[i].aspect = aspect
	end

	return _resolutionOptions
end

function love.load()
	lgr.setDefaultFilter("nearest", "nearest")
	lgr.setLineStyle("rough")

	DEBUGMODE = false

	preferences = lfs.load("preferences.lua")()
	LAST_MAP_LOADED_IN_EDITOR = preferences.lastEditedMap or ""
	THREEDEERESOLUTION = preferences.threeDeeResolution or 6

	appCursors = {
		normal = love.mouse.newCursor("images/cursors/normal.png", 2, 2),
		pencil = love.mouse.newCursor("images/cursors/pencil.png", 2, 2),
		line = love.mouse.newCursor("images/cursors/line.png", 2, 2),
		circle = love.mouse.newCursor("images/cursors/circle.png", 2, 2),
		rectangle = love.mouse.newCursor("images/cursors/rectangle.png", 2, 2),
		dropper = love.mouse.newCursor("images/cursors/dropper.png", 4, 2),
		marquee = love.mouse.newCursor("images/cursors/marquee.png", 2, 2),
		add = love.mouse.newCursor("images/cursors/add.png", 2, 2),
		remove = love.mouse.newCursor("images/cursors/remove.png", 2, 2),
		pan = love.mouse.newCursor("images/cursors/pan.png", 20, 20),
		erase = love.mouse.newCursor("images/cursors/erase.png", 2, 2),
		eraseall = love.mouse.newCursor("images/cursors/eraseall.png", 2, 2),
		move_horiz = love.mouse.newCursor("images/cursors/move_horiz.png", 18, 14),
		move_vert = love.mouse.newCursor("images/cursors/move_vert.png", 14, 18),
	}

	FLOORLAYER = -10000
	FLOOROBJLAYER = -9999.5
	SHADOWLAYER = -9990
	WALLLAYER = 0
	WALLDECOLAYER = 0.0001
	REFLECTIONLAYER = FLOORLAYER - 15

	RENDERWIDTH = 380
	RENDERHEIGHT = 260

	love.slowdownMultiplier = 1

	TARGETNIL = -100000000

	SHADOWCOLOR = {0,0,0,0.39}

	input = require 'library.input'
	input:init()
	bump = require 'library.bump'
	flux = require 'library.flux'
	cron = require 'library.cron'
	performanceGraph = require 'library.performancegraph'
	stack = require "library.stateStack"
	anim8 = require 'library.anim8'

	-- require "library.threeDeeWorld"
	require 'library.dumper'
	require 'library.functions'
	require 'library.entitySetup'

	require 'library.pathfinder'
	require 'behaviors'
	require 'hitbox'

	require 'states.setupScreen'
	-- require 'states.baseState'
	require 'states.vidResOptions'
	-- require (CUR_GAME..'.states.title')

	WINDOW_WIDTH, WINDOW_HEIGHT = lgr.getWidth(), lgr.getHeight()
	win.setTitle("Untitled Adventure Engine - F1: Debug  F3 Collisions")

	resizeScreen(1280, 720)

	stateStack = stack:init { drawWidth = RENDERWIDTH, drawHeight = RENDERHEIGHT, name = "Stack 1" }
	stateStack.useShader = preferences.scanlinesEnabled

	font = {
		dialog = lgr.newImageFont("fonts/dialog.png", " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`_*#=[]'{}|<>\"", 1),
		dialogbold = lgr.newImageFont("fonts/dialogbold.png", " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`_*#=[]'{}|<>", 1),
		tiny = lgr.newImageFont("fonts/tiny.png", " 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-,!:()[]{}<>\"'/=_|\\+%", 1),
		tinybordered = lgr.newImageFont("fonts/tinybordered.png", " 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-,!:()[]{}<>\"'/=_|\\+%", -1),
		large = lgr.newImageFont("fonts/large.png", " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`_*#=[]'{}|<>", 1),
		cga = lgr.newImageFont("fonts/cga.png", " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~", 1),
		mono2 = lgr.newImageFont("fonts/mono2.png", " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.,\"!?()_/-+:;\\<>", 1),
		mono1 = lgr.newImageFont("fonts/mono1.png", " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~…", 1),
		symbol = lgr.newImageFont("fonts/symbol.png", " 0123456789$KhjklH;", 1),
		symbolbordered = lgr.newImageFont("fonts/symbolbordered.png", " 0123456789.$KBhjklH;", -1),
		bordered = lgr.newImageFont("fonts/bordered.png", " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`_*#=[]'{}|<>$", -1),
	}
	font.dialog:setLineHeight(1.6)
	font.tiny:setLineHeight(0.75)
	-- font.tiny:setLineHeight(0.75)

	stateStack:push(setupScreen)

	love.resize()

	lgr.setBackgroundColor(0,0,0)
end

local xbl, xbu, ybl, ybu
function love.update(dt)
	WINDOW_WIDTH, WINDOW_HEIGHT = lgr.getWidth(), lgr.getHeight()

	-- dt = math.max(dt, 1/120)-- * love.slowdownMultiplier

	performanceGraph:preUpdate()

	if win.hasFocus() then
		flux.update(dt)
		cron.update(dt)
		stateStack:update(dt)
	end

	performanceGraph:postUpdate()
end

function love.draw()
	performanceGraph:preDraw()

	stateStack:draw()

	-- lgr.push()
	-- lgr.scale(2 * win.getDPIScale())
	-- lgr.translate(4, 4)
	-- lgr.setFont(font.dialog)
	-- lgr.setColor(0,0,0,1)
	-- lgr.rectangle("fill", -2, -2, 60, 19)
	-- lgr.setColor(1,1,1)
	-- lgr.print(tmr.getFPS(), 0, 0)
	-- lgr.print(math.floor(gameState.camera.x) .. "," .. math.floor(gameState.camera.y), 0, 8)
	-- lgr.pop()

	drawResourceUsage()

	if SHOWCAMERAREGION then
		lgr.setLineWidth(4)
		lgr.push()
		lgr.translate(20, 20)
		lgr.scale(1)

		-- lgr.setColor(0.4,0,0.4,0.54)
		-- lgr.rectangle("fill", 0, 0, gameState.camera.moveRegionWidth + RENDERWIDTH, gameState.camera.moveRegionHeight + RENDERHEIGHT)--(gameState.camera.rb - gameState.camera.lb), (gameState.camera.bb - gameState.camera.tb))

		local s = ((gameState.mapGridImage:getWidth() * TILESIZE) / (gameState.camera.moveRegionWidth + RENDERWIDTH) * TILESIZE) / 16
		lgr.setColor(1,1,1)
		lgr.draw(gameState.mapGridImage, 0, 0, 0, s, s)

		for i, r in pairs(gameState.cameraRegions) do
			if r.name ~= "default" then
				lgr.setColor(1,1,1)
				lgr.rectangle("line", r[1], r[2], r[3] - r[1], r[4] - r[2])
				lgr.setFont(font.tinybordered)

				local t = i .. " " .. r.name .. "\n"
				t = t .. r[1] .. "," .. r[2] .. "\n"
				t = t .. r[3] .. "," .. r[4] .. "\n"

				-- local w = font.bordered:getWidth(t)
				-- local h = font.bordered:getHeight(t) * font.bordered:getLineHeight()

				lgr.push()
				lgr.translate(r[1] + 4, r[2] + 4)
				lgr.scale(2)
				-- lgr.setColor(0,0,0,0.5)
				-- lgr.rectangle("fill", -1, -1, w + 4, h + 4)
				lgr.setColor(0,0,0,0.5)
				lgr.print(t, 1, 1)
				lgr.setColor(1,1,1)
				lgr.print(t, 0, 0)
				lgr.pop()
				-- lgr.line(r[1], r[2], r[1], r[4] - r[2])
				-- lgr.line(r[1], r[2], r[3] - r[1], r[2])
			end
		end

		lgr.setColor(1,1,1,0.6)
		lgr.push()
		lgr.translate(-gameState.camera.x, -gameState.camera.y)
		lgr.rectangle("fill", 0, 0, RENDERWIDTH, RENDERHEIGHT)

		lgr.push()
		lgr.scale(2)
		lgr.translate(2, 2)
		lgr.setFont(font.tinybordered)
		local t = "Camera: " .. -math.floor(gameState.camera.x) .. ", " .. -math.floor(gameState.camera.y) .. "\n"
		t = t .. "Player: " .. math.floor(gameState.gameEntities.player.x) .. ", " .. math.floor(gameState.gameEntities.player.y) .. "\n"
		if gameState.gameEntities.player.tileX and gameState.gameEntities.player.tileY then
			t = t .. "P Tile: " .. math.floor(gameState.gameEntities.player.tileX) .. ", " .. math.floor(gameState.gameEntities.player.tileY)
		end
		lgr.setColor(0,0,0,0.5)
		lgr.print(t, 1, 1)
		lgr.setColor(1,1,0)
		lgr.print(t, 0, 0)
		-- lgr.print("L:" .. gameState.camera.lb .. " R:" .. gameState.camera.rb, 0, 8)
		-- lgr.print("T:" .. gameState.camera.tb .. " B:" .. gameState.camera.bb, 0, 16)
		lgr.pop()

		lgr.pop()

		lgr.setColor(0,1,0)
		lgr.rectangle("fill", gameState.gameEntities.player.x, gameState.gameEntities.player.y, gameState.gameEntities.player.w, gameState.gameEntities.player.h)
		lgr.pop()
		lgr.setLineWidth(1)
	end

	performanceGraph:postDraw()

	if stateStack:getTopContext().drawDebugInfo then
		lgr.push()
		lgr.scale(2)
		lgr.translate(2, 2)
		stateStack:getTopContext():drawDebugInfo()
		lgr.pop()
	end

	-- input:draw()
end



function love.keypressed(k, u)
	input:keypressed(k, r)
	stateStack:keypressed(k, r)

	if k == "f1" then
		performanceGraph:toggle()
	elseif k == "f3" then
		DEBUGMODE = not DEBUGMODE
	elseif k == "f5" then
		love.event.quit("restart")
	elseif k == "f6" then
		stateStack.useShader = not stateStack.useShader
	elseif k == "f7" then
		DEBUGPRINT = not DEBUGPRINT
	end
end

function love.keyreleased(k, r)
	input:keyreleased(k, r)
	stateStack:keyreleased(k, r)
end

function love.textinput(k)
	input:textinput(k)
	stateStack:textinput(k)
end



function love.focus(f)
	stateStack:focus(f)
end

function love.quit()
	print("Quitting", LAST_MAP_LOADED_IN_EDITOR)
	-- if PREFS_CHANGED then
		preferences = {
			renderwidth = RENDERWIDTH,
			renderheight = RENDERHEIGHT,
			scanlinesEnabled = stateStack.useShader,
			lastEditedMap = LAST_MAP_LOADED_IN_EDITOR,
			threeDeeResolution = THREEDEERESOLUTION,
			gameSeason = GAMESEASON,
		}
		local p = DataDumper(preferences, nil, false)
		lfs.write("preferences.lua", p, all)
	-- end

	if gameState then
		local p = DataDumper(gameState.chestIDList, nil, false)
		lfs.write("chests.lua", p, all)
	end
end

function love.resize(w, h)
	WINDOW_WIDTH, WINDOW_HEIGHT = lgr.getWidth(), lgr.getHeight()

	local w = w or WINDOW_WIDTH
	local h = h or WINDOW_HEIGHT
	print("Resized window:", w / PIXELSCALE, h / PIXELSCALE)

	local sw, sh = w / screen.drawWidth, h / screen.drawHeight

	local aspect_game = w / h
	local aspect_window = screen.drawWidth / screen.drawHeight

	if aspect_window <= aspect_game then
		screen.drawScale = sh
	else
		screen.drawScale = sw
	end

	local ho = stateStack.drawHorizOffset / PIXELSCALE
	local vo = stateStack.drawVertOffset / PIXELSCALE
	-- screen.overlayPixelSize = 2
	-- if WINDOW_HEIGHT <= 500 then
	-- 	screen.overlayPixelSize = 1
	-- end
	-- screen.overlay = lgr.newCanvas((w / PIXELSCALE) / screen.overlayPixelSize - ho, (h / PIXELSCALE) / screen.overlayPixelSize - vo)

	stateStack:resize(w, h)

	if editorState then
		if editorState.loaded then
			resizeEditorWindowDimensions(editorState, w, h)
		elseif scriptEditor.loaded then
			resizeEditorWindowDimensions(scriptEditor, w, h)
		end
	end
end



function love.visible(v)
	print(v and "Window is visible!" or "Window is not visible!")
end



function love.mousefocus(f)
	if not f then
		print("LOST MOUSE FOCUS")
	else
		print("GAINED MOUSE FOCUS")
	end
end

function love.mousemoved(x, y, dx, dy)
	stateStack:mousemoved(x, y, dx, dy)
end

function love.mousepressed(x, y, button)
	stateStack:mousepressed(x, y, button)
end

function love.mousereleased(x, y, button)
	if timeSinceLastMouserelease > -1 and tmr.getTime() - timeSinceLastMouserelease < 0.5 and pointInRectangle(x, y, lastMouseClickX - 2, lastMouseClickY - 2, 4, 4) then
		__dprint("Double-click")
		timeSinceLastMouserelease = -1
		stateStack:mousereleased(x, y, button)
		stateStack:doubleclick(x, y, button)
	else
		__dprint("Click")
		timeSinceLastMouserelease = tmr.getTime()
		lastMouseClickX = x
		lastMouseClickY = y
		stateStack:mousereleased(x, y, button)
	end
end

function love.wheelmoved(x, y)
	stateStack:wheelmoved(x, y)
end




function love.gamepadaxis(joystick, axis, value)
	input:gamepadaxis(joystick, axis, value)
end

function love.gamepadpressed(joystick, button)
	input:gamepadpressed(joystick, button)
end

function love.gamepadreleased(joystick, button)
	input:gamepadreleased(joystick, button)
end

function love.joystickadded(joystick)
	input:joystickadded(joystick)
end

function love.joystickaxis(joystick, axis, value)
	input:joystickaxis(joystick, axis, value)
end

function love.joystickhat(joystick, hat, direction)
	input:joystickhat(joystick, hat, direction)
end

function love.joystickpressed(joystick, button)
	input:joystickpressed(joystick, button)
end

function love.joystickreleased(joystick, button)
	input:joystickreleased(joystick, button)
end

function love.joystickremoved(joystick)
	input:joystickremoved(joystick)
end


function resizeEditorWindowDimensions(e, w, h)
	e = e or editorState
	local ps = e.pixelSize * win.getDPIScale()
	local w = w or lgr.getWidth()
	local h = h or lgr.getHeight()
	if w <= 640 then ps = 1 * win.getDPIScale() end
	print(">>", w / 3, h / 3)
	stateStack:changeDrawDimensions(w / ps, h / ps)
end

function getGameObject(name)
	local g = gameState.gameEntities[name]
	return g
end

function drawDebugBoundingBox(state, e, t)
	local tm = tmr.getTime()
	if DEBUGMODE then
		local t = t or ""
		-- t = t .. "\n" .. e.cameraRegion or "nil"
		if not state.collisionMap:hasItem(e) then return false end
		local x,y,w,h = state.collisionMap:getRect(e)

		if e.kind == "hitbox" then
			state.drawPool:push { kind = "rect", fill = "line", color = e._dbgColor or {1,0,0,0.78}, x = x + 1, y = y + 1, w = w - 1, h = h - 1, layer = 12001 }
			state.drawPool:push { kind = "text", text = "HP:" .. (e.owner.hp or 0), color = {1,1,1}, font = font.tiny, x = x, y = y - 6, layer = 12003 }
		else
			state.drawPool:push { kind = "rect", fill = "fill", color = e._dbgColor or (e.colType == "world" and {1,150,0.39,0.78} or {0.78,0.78,0,0.39}), x = x, y = y, w = w, h = h, layer = 10000 }
			state.drawPool:push { kind = "rect", fill = "line", color = e._dbgColor or {1,1,1,0.39}, x = x + 1, y = y + 1, w = w - 1, h = h - 1, layer = 10001 }

			if e._dbgText then
				t = t .. "\n" .. e._dbgText
			end

			state.drawPool:push { kind = "text", text = t, color = {1,1,1}, font = font.tinybordered, x = x, y = y, layer = 10003 }
		end

		if e._targetX ~= TARGETNIL and e._targetY ~= TARGETNIL and e._targetX and e._targetY and e.cx and e.cy then
			local c = {1,1,1,0.39}
			if e.master == "player" then c = {1,0,0,0.78} end
			state.drawPool:push { kind = "line", points = { e._targetX, e._targetY, e.cx, e.cy }, lineWidth = 1, color = c, layer = 10100 }
			state.drawPool:push { kind = "circle", x = e._targetX, y = e._targetY, rad = 2, color = c, layer = 10099 }
			-- state.drawPool:push { kind = "text", text = tostring(e._targetX) .. "," .. tostring(e._targetY), color = {1,1,1}, font = font.tiny, x = e._targetX, y = e._targetY, layer = 10003 }
		end

		if e.direction and e._targetX ~= TARGETNIL and e.cx and e.cy then
			state.drawPool:push { kind = "line", points = { e.cx, e.cy, e.cx + math.cos(e.direction) * e.speed, e.cy + math.sin(e.direction) * e.speed }, lineWidth = 1, color = {0,1,0}, layer = 10101 }
		end

	end
	performanceGraph:addTodebugDrawTime(tmr.getTime() - tm)
end

function resizeScreen(w, h, highdpi)
	local w = w or lgr.getWidth()
	local h = h or lgr.getHeight()
	local s = 2
	if highdpi then dpis = 2 end
	screen = {
		drawWidth = w, --960
		drawHeight = h, --540
		simpleMode = false,
		overallScale = s,
		debugTextScale = dpis or 1,
		focused = false
	}
	-- screen.zoomScale = 4.5 * (screen.drawWidth / 1920)
	screen.zoomScale = 1
	PIXELSCALE = love.window.getDPIScale()
end


function loadAnimationForImage(self, loc, image, anim, onLoop)
	__dprint("Load animation for " .. loc .. " " .. image)
	self._animations = self._animations or {}
	self._animations[image] = self._animations[image] or {}
	local img = imgLib:image(image)
	local w, h = img:getWidth(), img:getHeight()
	local f = CUR_GAME..'/images/'..loc..'/'..image..'_anim.lua'
	if lfs.getInfo(f) then
		local data = lfs.load(f)()
		local g = anim8.newGrid(data.grid[1], data.grid[2], w, h)
		if anim then
			__dprint(" > Setup:", image, anim)
			local a = data.anims[anim]
			-- if type(a[3]) == "string" then onLoop = a[3] end
			a[3] = a[3] or nil
			self._animations[image][anim] = anim8.newAnimation(g(unpack(a[1])), a[2], onLoop)
			if a[3] then
				self._animations[image][anim]:setPalette(a[3])
			end
		else
			for i, a in pairs(data.anims) do
				__dprint(" > Setup:", image, i)
				-- if type(a[3]) == "string" then onLoop = a[3] end
				a[3] = a[3] or nil
				self._animations[image][i] = anim8.newAnimation(g(unpack(a[1])), a[2], onLoop)
				if a[3] then
					self._animations[image][i]:setPalette(a[3])
				end
			end
		end
	else
		__dprint(" > No ANIM file available for " .. image)
		local data = { grid = { 16, 16 }, anims = {} }
		local mw, mh = math.floor(w / 16), math.floor(h / 16)
		local x, y = 1, 1
		local done = false
		while not done do
			data.anims[#data.anims+1] = { {x, y}, 1 }
			x = x + 1
			if x > mw then
				x = 1
				y = y + 1
				if y > mh then
					done = true
				end
			end
		end

		local g = anim8.newGrid(data.grid[1], data.grid[2], w, h)
		for i, a in pairs(data.anims) do
			__dprint(" > Setup:", image, i)
			self._animations[image][i] = anim8.newAnimation(g(unpack(a[1])), a[2], onLoop)
		end
	end
end

function updateAnimation(self, image, anim)

end


function drawResourceUsage()
	if performanceGraph.enabled then

		lgr.push()

		if DEBUGENTITYLIST then
			lgr.push()
			lgr.scale(2 * PIXELSCALE)
			lgr.translate(4, 4)

			local f = font.tinybordered
			lgr.setFont(f)
			local t, l = f:getWrap(DEBUGENTITYLIST, 1000)
			local w, h = t, #l * f:getLineHeight() * f:getHeight() - 1
			-- lgr.setColor(0,0,0,0.79)
			-- lgr.rectangle("fill", -3, -2, w + 6, h + 4)

			-- lgr.setColor(0,0,0)
			-- lgr.printf(DEBUGENTITYLIST or "", 1, 1, 512, "left")
			lgr.setColor(1,1,1)
			lgr.printf(DEBUGENTITYLIST or "", 0, 0, 256, "left")
			lgr.pop()
		end

		lgr.translate(lgr.getWidth() - 85 * 2 * PIXELSCALE, lgr.getHeight() - 48 * PIXELSCALE)

		lgr.push()
		lgr.scale(2 * PIXELSCALE)

		local m = math.floor(collectgarbage("count")/1024)
		local tm = lgr.getStats()
		-- lgr.printf("Mem: " .. m .. "MB / FPS: "..tmr.getFPS(), 0, 0, 200, "right")
		-- lgr.printf(string.format("Tex Mem: %.2f MB", tm.texturememory / 1024 / 1024), 0, 8, 200, "right")
		-- lgr.printf("Canvas Count: "..tm.canvases, 0, 16, 200, "right")
		-- lgr.printf("Draw Calls: "..tm.drawcalls, 0, 24, 200, "right")
		-- lgr.setColor(1,1,1,0.4)
		-- lgr.rectangle("fill", -m + 200, -3, m, 1)

		lgr.setColor(0,0,0)
		lgr.rectangle("fill", -m + 84, -4, m + 2, 3)
		lgr.setColor(1,1,1)
		lgr.rectangle("fill", -m + 85, -3, m, 1)

		local t1 = "Memory:\nTex Mem:\nCanvases:\nDraw Calls:"
		local t2 = m .. "MB / FPS: "..tmr.getFPS() .. "\n" .. string.format("%.2f MB", tm.texturememory / 1024 / 1024) .. "\n" .. tm.canvases .. "\n" .. tm.drawcalls
		lgr.setFont(font.tinybordered)
		lgr.printf(t1, 0, 0, 85, "left")
		lgr.printf(t2, 0, 0, 85, "right")


		lgr.pop()

		lgr.pop()

		lgr.push()
		lgr.translate(lgr.getWidth() - 85 * 2 * PIXELSCALE, 2)--RENDERHEIGHT * stateStack.drawScale + 2 * PIXELSCALE)

		lgr.push()
		lgr.scale(2 * PIXELSCALE)
		lgr.setFont(font.tinybordered)
		lgr.print("Resolution: ".. RENDERWIDTH .. " x " .. RENDERHEIGHT .. "\nTile Grid: " .. math_round(RENDERWIDTH / TILESIZE, 2) .. " x " .. math_round(RENDERHEIGHT / TILESIZE, 2), 0, 0)
		if gameState then
			if gameState.drawPool then
				lgr.translate(0, 16)
				lgr.print("Draw Pool:", 0, 0)
				lgr.translate(0, 6)
				local cnt = gameState.drawPool:getCounts()
				local y = 0
				for i, c in pairs(cnt) do
					lgr.print(i..":", 0, y)
					lgr.printf(c, 0, y, 85, "right")
					y = y + 6
				end
			end
		end

		lgr.pop()

		lgr.pop()
	end
end
